# Exceptions can be put here to get them out of the way of the main flow of things,
# and because once in a while we may want them to be shared or to have shared parents.

from dcicutils.misc_utils import full_object_name, full_class_name, capitalize1


class KnownBugError(AssertionError):

    def set_jira_ticket(self, jira_ticket):
        self.jira_ticket = jira_ticket
        self.bug_name = "a bug" if jira_ticket is None else "Bug %s" % jira_ticket


class UnfixedBugError(KnownBugError):
    pass


class WrongErrorSeen(UnfixedBugError):

    def __init__(self, *, expected_class, error_seen, jira_ticket=None):
        self.set_jira_ticket(jira_ticket)
        self.expected_class = expected_class
        self.error_seen = error_seen
        super().__init__("%s did not occur where expected."
                         " An error of class %s was thrown where one of class %s was expected."
                         " The bug may have been fixed or it may be manifesting in an unexpected way: %s"
                         % (capitalize1(self.bug_name),
                            full_class_name(error_seen),
                            full_object_name(expected_class),
                            error_seen))


class ExpectedErrorNotSeen(UnfixedBugError):

    def __init__(self, *, jira_ticket=None):
        self.set_jira_ticket(jira_ticket)
        super().__init__("%s did not occur where expected. It may have been fixed." % capitalize1(self.bug_name))


class FixedBugError(KnownBugError):
    pass


class WrongErrorSeenAfterFix(FixedBugError):

    def __init__(self, *, expected_class, error_seen, jira_ticket=None):
        self.set_jira_ticket(jira_ticket)
        self.expected_class = expected_class
        self.error_seen = error_seen
        super().__init__("%s did not occur,"
                         " but an error of class %s was thrown where one of class %s was expected to be fixed."
                         " This may be another bug showing through or a regression of some sort: %s"
                         % (capitalize1(self.bug_name),
                            full_class_name(error_seen),
                            full_object_name(expected_class),
                            error_seen))


class UnexpectedErrorAfterFix(FixedBugError):

    def __init__(self, *, expected_class, error_seen, jira_ticket=None):
        self.set_jira_ticket(jira_ticket)
        self.expected_class = expected_class
        self.error_seen = error_seen
        super().__init__("An error of class %s was thrown where we expected one of class %s was fixed."
                         " This may be a regression in the fix for %s: %s"
                         % (full_class_name(error_seen), full_object_name(expected_class), self.bug_name, error_seen))



import boto3
from .misc_utils import PRINT


def get_ecs_real_url(env):
    """ Grabs the ECS target URL created on orchestration from Cloudformation.
        NOTE: this is not intended to function with multiple ECS stacks in the
        same account as of right now.
    """
    cfn_client = boto3.client('cloudformation')
    stacks = cfn_client.describe_stacks().get('Stacks', [])
    for stack in stacks:
        for output in stack.get('Outputs', []):
            if output.get('OutputKey', '') == ('ECSApplicationURL%s' % env.replace('-', '')):
                return output.get('OutputValue')
    PRINT('Did not locate the server from Cloudformation! Check ECS Stack metadata.')
    return ''


def get_ecr_repo_url(env):
    """ Gets the ECR repository URL of the repository from which ECS pulls application
        image runtimes from.
    """
    cfn_client = boto3.client('cloudformation')
    stacks = cfn_client.describe_stacks().get('Stacks', [])
    for stack in stacks:
        for output in stack.get('Outputs', []):
            if output.get('OutputKey', '') == 'C4EcrTrialAlphaECRRepoURL':  # XXX: hardcoded constant from 4dn-cloud-infra
                val = output.get('OutputValue')
                if env in val:
                    return val
    PRINT('Did not locate the server from Cloudformation! Check ECS Stack metadata.')
    return ''



import boto3
from .misc_utils import PRINT
from .ecr_utils import CGAP_ECR_REGION as CGAP_ECS_REGION


class ECSUtils:
    """ Utility class for interacting with ECS - mostly stubs at this point, but will likely
        expand a lot.
    """

    def __init__(self):
        """ Creates a boto3 client for 'ecs'. """
        self.client = boto3.client('ecs', region_name=CGAP_ECS_REGION)  # same as ECR

    def list_ecs_clusters(self):
        """ Returns a list of ECS clusters ARNs. """
        return self.client.list_clusters().get('clusterArns', [])

    def list_ecs_services(self, *, cluster_name):
        """ Returns a list of ECS Service names under the given cluster. There should be 4 of them as of right now -
            adding more is considered a compatible change.
        """
        services = self.client.list_services(cluster=cluster_name).get('serviceArns', [])
        if len(services) < 4:
            raise Exception('Environment error! Expected at least 4 services - check that ECS finished orchestrating.')
        return services

    def update_ecs_service(self, *, cluster_name, service_name):
        """ Forces an update of this cluster's service, by default WSGI (for now)
            It is highly likely we want some variants of this - but will start here.
            Note that this just updates a particular service.
        """
        try:
            self.client.update_service(cluster=cluster_name, service=service_name,
                                       forceNewDeployment=True)
            PRINT('Successfully updated ECS cluster %s service %s' % (cluster_name, service_name))
        except Exception as e:
            PRINT('Error encountered triggering cluster update: %s' % e)
            raise

    def update_all_services(self, *, cluster_name):
        """ Forces an update of all services on this cluster. """
        for service_name in self.list_ecs_services(cluster_name=cluster_name):
            self.update_ecs_service(cluster_name=cluster_name, service_name=service_name)
        return True
